type const = string
type var   = string

type t =
  | Type
  | Var of var
  | Prod of var * t * t
  | Lam  of var * t * t
  | App  of t * (t * t)

type context = (var * t) list

type fresh_seed = int

let domain   gamma = List.map fst gamma
let codomain gamma = List.map snd gamma

let gensym : fresh_seed -> var * fresh_seed =
  fun seed ->
    ("v"^(string_of_int seed), seed+1)

let rec fv gamma expr acc =
  match expr with
  | Type -> acc
  | Var v -> 
      if List.mem v gamma then acc 
      else (v :: acc)
  | Prod(x,tx,e) ->
      fv (x::gamma) e (fv gamma tx acc)
  | Lam(x,tx,e) ->
      fv (x::gamma) e (fv gamma tx acc)
  | App(f,a) ->
      fv gamma f (fv gamma a acc)

let rec fvsubst oldvar newvar expr =
  match expr with
  | Type -> Type
  | Var v ->
      if v = oldvar then
	Var newvar
      else
	expr
  | Prod(x,tx,e) ->
      if x = oldvar then
	expr
      else
	Prod(x,
	     fvsubst oldvar newvar tx,
	     fvsubst oldvar newvar e)
  | Lam(x,tx,e) ->
      if x = oldvar then
	expr
      else
	Prod(x,
	     fvsubst oldvar newvar tx,
	     fvsubst oldvar newvar e)
  | App(f,(a,ta)) ->
      App(fvsubst oldvar newvar f,
	  (fvsubst oldvar newvar a,
	   fvsubst oldvar newvar ta))

(* subst fresh var expr1 expr2 = fresh, expr2[var\expr1] *)
let subst fresh var expr1 expr2 =
  let e1fv = fv [] expr1 [] in
  let rec subst fresh expr1 expr2 =
    match expr2 with
    | Type -> fresh, Type
    | Var v ->
	if v = var then
	  fresh, expr1
	else
	  fresh, expr2
    | Prod(x,tx,e) ->
	let fresh, expr1' =
	  if List.mem x e1fv then
	    let newvar,fresh = gensym fresh in
	    fresh, (fvsubst x newvar expr1)
	  else
	    fresh, expr1
	in
	if var = x then
	  let fresh,tx' = subst fresh expr1 tx in
	  fresh, (Prod(x,tx',e))
	else
	  let fresh,tx' = subst fresh expr1 tx in
	  let fresh,e'  = subst fresh expr1' e in
	  fresh, (Prod(x,tx',e'))
    | Lam(x,tx,e) ->
	let fresh, expr1' =
	  if List.mem x e1fv then
	    let newvar,fresh = gensym fresh in
	    fresh, (fvsubst x newvar expr1)
	  else
	    fresh, expr1
	in
	if var = x then
	  let fresh,tx' = subst fresh expr1 tx in
	  fresh, (Prod(x,tx',e))
	else
	  let fresh,tx' = subst fresh expr1 tx in
	  let fresh,e'  = subst fresh expr1' e in
	  fresh, (Prod(x,tx',e'))
    | App(f,(a,ta)) ->
	let fresh,f'  = subst fresh expr1 f in
	let fresh,a'  = subst fresh expr1 a in
	let fresh,ta' = subst fresh expr1 ta in
	fresh, (App(f',(a',ta')))
  in
  subst fresh expr1 expr2

let rec valid_kind gamma kind =
  match kind with
  | Prod(x,a,k) ->
      (typecheck gamma a Type)
      && (valid_kind ((x,a) :: gamma) k)
  | Type ->
      (match gamma with
       | [] -> true
       | (x,a) :: gamma' ->
	   (* Optimisation possible ici, en regardant la geule de a ? *)
	   (not (List.mem x (domain gamma')))
	   && ((valid_family gamma' a Type) || (valid_kind gamma a)))
  | _ -> false

and valid_family gamma expr kind =
  match expr with
  | Var x ->
      (* Attention pour le List.mem : il faut prendre en compte l'alpha-convertibilit. TODO *)
      (valid_kind gamma Type) && (List.mem (x,kind) gamma)
  | Prod(x,a,b) ->
      if kind = Type then
	(valid_family gamma a Type) && (valid_family ((x,a) :: gamma)  b Type)
      else
	false
  | Lam(x,a,b) ->
      (match kind with
       | Prod(x,a',k) ->
	   if a = a' then
	     (valid_family gamma a Type) && (valid_family ((x,a) :: gamma) b Type)
	   else
	     false
       | _ -> false)
  | App(a,(m,tm)) ->
      
      
