%{

%}

%token ARROW FUNARROW
%token FUNCTION FORALL EXISTS PLUS LET EQUAL ASSIGN IN DOT PRJ1 PRJ2 CASE INJ1 INJ2 REFL SUBST
%token TYPE KIND CONSTANT DEFINE PIPE
%token OPENP CLOSEP
%token COMMA COLON EOF
%token <string> FLOAT INTEGER BOOL IDENT CONSTRUCTOR 

%type <Ast.concrete> term

%start program
%type <Ast.program> program


%%
program: 
  | const_definition program { Ast.Const($1,$2) }
  | term_definition program { Ast.Def($1,$2) }
  | EOF { Ast.End }
;

const_definition:
  | CONSTANT IDENT COLON term DOT { ($2,$4) }
;

term_definition:
  | DEFINE IDENT COLON term ASSIGN term DOT { ($2,$4,$6) }
;

terminal:
  | IDENT { Ast.CVar($1) }
  | KIND { Ast.CKind }
  | TYPE { Ast.CType }
;

safe:
  | terminal { $1 }
  | OPENP term CLOSEP { $2 }
;

typed_ident:
  | OPENP id=IDENT COLON t=term CLOSEP { (id,t) }

arrow:
  | safe ARROW arrow { Ast.CProd("_",$1,$3) }
  | safe ARROW safe { Ast.CProd("_",$1,$3) }
;

forall:
  | FORALL ti=typed_ident+ COMMA last=forall
  | FORALL ti=typed_ident+ COMMA last=exists
  | FORALL ti=typed_ident+ COMMA last=sum
  | FORALL ti=typed_ident+ COMMA last=equality
  | FORALL ti=typed_ident+ COMMA last=safe 
    { List.fold_right (fun ti acc -> Ast.CProd(fst ti,snd ti,acc)) ti last }
;

exists:
  | EXISTS ti=typed_ident COMMA last=forall
  | EXISTS ti=typed_ident COMMA last=exists
  | EXISTS ti=typed_ident COMMA last=sum
  | EXISTS ti=typed_ident COMMA last=equality
  | EXISTS ti=typed_ident COMMA last=safe { Ast.CSig(fst ti,snd ti,last) }
;

sum:
  | safe PLUS safe { Ast.CSum($1,$3) }
;

equality:
  | safe EQUAL safe { Ast.CEqu($1,$3) }
;

func:
  | FUNCTION ti=typed_ident+ FUNARROW last=func
  | FUNCTION ti=typed_ident+ FUNARROW last=apply
  | FUNCTION ti=typed_ident+ FUNARROW last=safe 
    { List.fold_right (fun ti acc -> Ast.CLam(fst ti,snd ti,acc)) ti last }
;

apply:
  | terminal safe { Ast.CApp($1,$2) }
  | OPENP term CLOSEP safe { Ast.CApp($2,$4) }
  | apply safe { Ast.CApp($1,$2) }
;


term:
  | terminal { $1 }
  | func { $1 }
  | arrow { $1 }
  | forall { $1 }
  | exists { $1 }
  | sum { $1 }
  | equality { $1 }
  | OPENP term COMMA term PIPE term CLOSEP { Ast.CPair($6,$2,$4) }
  | PRJ1 safe { Ast.CPrj1($2) }
  | PRJ2 safe { Ast.CPrj2($2) }
  | INJ1 safe IN safe { Ast.CInl($4,$2) }
  | INJ2 safe IN safe { Ast.CInr($4,$2) }
  | CASE safe safe safe { Ast.CCase($2,$3,$4) }
  | REFL safe { Ast.CRefl($2) }
  | SUBST safe safe safe safe safe { Ast.CSubst($2,$3,$4,$5,$6) }
  | OPENP term CLOSEP { $2 }
  | apply { $1 }
;

%%
