open Ast
open Printf
open Lf

let rec interpret fresh context =
  function
  | Const((ident, ast), tail) ->
      let term     = ast_to_lf context [] ast in
      let term     = normalize_cbn context term in
      let fresh, s = to_string fresh [] term in
      let _        = printf "%s : %s\n" ident s in
      let fresh, _ = typecheck fresh context term in
	interpret 
	  fresh
	  (context_add_var context ident term)
	  tail

  | Def((ident, typ, ast), tail) ->
      let typ         = ast_to_lf context [] typ in
      let fresh, _    = typecheck fresh context typ in
      let term        = ast_to_lf context [] ast in
      let fresh, typ' = typecheck fresh context term in
      let term        = normalize_cbn context term in
      let typ         = normalize_cbn context typ in
      let fresh, s'   = to_string fresh [] term in
      let fresh, t'   = to_string fresh [] typ' in
	if typ = typ' then
	  let _ = printf "%s : %s proved.\n" ident t' in
	    interpret 
	      fresh
	      (context_add_def context ident term typ)
	      tail 
	else
	  printf "Incorrect proof of type %s.\n" t'

  | End -> ()


let input = Lexing.from_channel (open_in Sys.argv.(1))

let program = Parser.program Lf_lexer.token input

let _ = interpret 0 empty_context program
